/*
 * Copyright 2002-present the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.web.accept;

import java.util.List;

import org.jspecify.annotations.Nullable;
import org.junit.jupiter.api.Test;

import org.springframework.web.testfixture.servlet.MockHttpServletRequest;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

/**
 * Unit tests for {@link DefaultApiVersionStrategy}.
 * @author Rossen Stoyanchev
 */
public class DefaultApiVersionStrategiesTests {

	private static final SemanticApiVersionParser parser = new SemanticApiVersionParser();

	private final MockHttpServletRequest request = new MockHttpServletRequest();


	@Test
	void defaultVersionIsParsed() {
		String version = "1.2.3";
		ApiVersionStrategy strategy = apiVersionStrategy(version);
		assertThat(strategy.getDefaultVersion()).isEqualTo(parser.parseVersion(version));
	}

	@Test
	void validateSupportedVersion() {
		String version = "1.2";
		DefaultApiVersionStrategy strategy = apiVersionStrategy();
		strategy.addSupportedVersion(version);
		validateVersion(version, strategy);
	}

	@Test
	void rejectUnsupportedVersion() {
		assertThatThrownBy(() -> validateVersion("1.2", apiVersionStrategy()))
				.isInstanceOf(InvalidApiVersionException.class)
				.hasMessage("400 BAD_REQUEST \"Invalid API version: '1.2.0'.\"");
	}

	@Test
	void validateDetectedVersion() {
		String version = "1.2";
		DefaultApiVersionStrategy strategy = apiVersionStrategy(null, true);
		strategy.addMappedVersion(version);
		validateVersion(version, strategy);
	}

	@Test
	void validateWhenDetectedVersionOff() {
		String version = "1.2";
		DefaultApiVersionStrategy strategy = apiVersionStrategy();
		strategy.addMappedVersion(version);
		assertThatThrownBy(() -> validateVersion(version, strategy)).isInstanceOf(InvalidApiVersionException.class);
	}

	@Test
	void missingRequiredVersion() {
		assertThatThrownBy(() -> validateVersion(null, apiVersionStrategy()))
				.isInstanceOf(MissingApiVersionException.class)
				.hasMessage("400 BAD_REQUEST \"API version is required.\"");
	}

	private static DefaultApiVersionStrategy apiVersionStrategy() {
		return apiVersionStrategy(null, false);
	}

	private static DefaultApiVersionStrategy apiVersionStrategy(@Nullable String defaultVersion) {
		return apiVersionStrategy(defaultVersion, false);
	}

	private static DefaultApiVersionStrategy apiVersionStrategy(
			@Nullable String defaultVersion, boolean detectSupportedVersions) {

		return new DefaultApiVersionStrategy(
				List.of(request -> request.getParameter("api-version")),
				new SemanticApiVersionParser(), true, defaultVersion, detectSupportedVersions, null);
	}

	private void validateVersion(@Nullable String version, DefaultApiVersionStrategy strategy) {
		strategy.validateVersion(version != null ? parser.parseVersion(version) : null, request);
	}

}
